
/** This version of ExifViewer supports JPEG files only **/

import java.awt.Graphics2D;
import java.awt.RenderingHints;
import javax.imageio.ImageIO;
import javax.swing.JFrame;
import javax.swing.JButton;
import java.awt.event.ActionListener;
import java.awt.event.ActionEvent;
import javax.swing.table.DefaultTableModel;
import javax.swing.table.TableColumn;
import javax.swing.ImageIcon;
import javax.swing.JFileChooser;
import javax.swing.JLabel;
import javax.swing.JOptionPane;
import javax.swing.JSeparator;
import javax.swing.JScrollPane;
import java.awt.image.BufferedImage;
import java.io.ByteArrayInputStream;
import java.io.File;
import java.io.FileNotFoundException;
import java.io.FileOutputStream;
import java.io.FilenameFilter;
import java.io.IOException;
import java.io.InputStream;
import javax.swing.JComboBox;
import javax.swing.JTable;
import java.io.*;
import java.util.*;
import org.apache.sanselan.*;
import org.apache.sanselan.common.*;
import org.apache.sanselan.formats.jpeg.*;
import org.apache.sanselan.formats.tiff.*;
import org.apache.sanselan.formats.tiff.constants.*;
import java.awt.Color;
import ij.*;
import ij.process.*;
import ij.gui.*;
import java.awt.*;
import ij.plugin.*;
import ij.plugin.frame.*;


public class Exif_Viewer implements PlugIn{
	
	private JFrame frmExifViewer;
	private JTable exifTable;
	private DefaultTableModel myModel;
	private static TiffImageMetadata exifMetadata = null;
	private static File selected = null;
	private double psnr = 0;
	private ImagePlus imp = null;
    
	/** Launch the plugin **/
	public void run(String arg) {
		Exif_Viewer window = new Exif_Viewer();
		window.frmExifViewer.setVisible(true);
	}
	
	/** Create application window **/
	public Exif_Viewer() {
		initialize();
	}

	
	/** Initialize the window content **/
	private void initialize() {
		
		
		//Window
		frmExifViewer = new JFrame();
		frmExifViewer.setTitle("Exif Viewer - by Daniele Caruso & Dario Messina");
		frmExifViewer.setResizable(false);
		frmExifViewer.setAlwaysOnTop(false);
		frmExifViewer.setBounds(100, 100, 470, 690);
		frmExifViewer.setDefaultCloseOperation(JFrame.DISPOSE_ON_CLOSE);
		frmExifViewer.getContentPane().setLayout(null);
		
		//Label Thumbnail
		final JLabel labelThumbnail = new JLabel("");
		labelThumbnail.setBounds(119, 180, 232, 170);
		frmExifViewer.getContentPane().add(labelThumbnail);
		
		//Text PSNR
		final JLabel textPSNR = new JLabel("");
		textPSNR.setBounds(66, 377, 444, 21);
		frmExifViewer.getContentPane().add(textPSNR);
						
		//File List
		final JComboBox comboBoxFileList = new JComboBox();
		comboBoxFileList.addItem("...");
		comboBoxFileList.setEnabled(false);
		comboBoxFileList.addActionListener(new ActionListener() {
			public void actionPerformed(ActionEvent arg0) {
				if(comboBoxFileList.getSelectedIndex()!=0){
					selected = (File) comboBoxFileList.getSelectedItem();
					fillExifTable(selected);
					if(psnr>=20){
                        textPSNR.setText("Class 1 - Image has high probability of being original!");
                        labelThumbnail.setIcon(new ImageIcon(selected.getParent()+"/Thumbnails/"+selected.getName().substring(0, selected.getName().length()-4)+"Thumb.jpg"));
                        labelThumbnail.repaint();
					}
					else if(psnr >= 0 && psnr <20) {
						textPSNR.setText("Class 2 – Image has high probability of being processed/edited!");
                        labelThumbnail.setIcon(new ImageIcon(selected.getParent()+"/Thumbnails/"+selected.getName().substring(0, selected.getName().length()-4)+"Thumb.jpg"));
                        labelThumbnail.repaint();
					}
                    else if(psnr == -1){
                        textPSNR.setText("Image doens't contain any thumbnail!");
                        labelThumbnail.setText("");
                        labelThumbnail.repaint();
                    }
                        
				}
				else{
					labelThumbnail.setIcon(null);
					labelThumbnail.repaint();
				}
			}
		});
		comboBoxFileList.setBounds(10, 109, 444, 21);
		frmExifViewer.getContentPane().add(comboBoxFileList);
		
		//Exif Data Panel
		JScrollPane scrollPaneExifData = new JScrollPane();
		scrollPaneExifData.setBounds(10, 441, 444, 218);
		frmExifViewer.getContentPane().add(scrollPaneExifData);
		
		//Exif Table
		myModel = new myTableModel();
		exifTable = new JTable(myModel);	
		TableColumn col0 = exifTable.getTableHeader().getColumnModel().getColumn(0);
		col0.setMinWidth(150);
		col0.setMaxWidth(200);
		exifTable.setSurrendersFocusOnKeystroke(true);
		exifTable.setBackground(Color.WHITE);
		scrollPaneExifData.setViewportView(exifTable);
		
		//Button Browse
		JButton buttonBrowse = new JButton("Browse");
		buttonBrowse.addActionListener(new ActionListener() {
			public void actionPerformed(ActionEvent arg0) {
				{
                    try
                    {
                      JFileChooser browser = new JFileChooser();	
                      browser.setFileSelectionMode(JFileChooser.FILES_AND_DIRECTORIES );
                      browser.showOpenDialog(null); 				 
                      File input = browser.getSelectedFile();	
                      comboBoxFileList.setEnabled(true);
                  	  if(input.isFile()){	//input is a file!
                  		  String exif = readMetaData(input);
                  		  psnr = extractThumbnail(input);
                          if(psnr != -2){
                              comboBoxFileList.addItem(input);
                              writeReport(input, exif, psnr);
                              JOptionPane.showMessageDialog(null,"Select an image from the pull-down menu","Done!",JOptionPane.INFORMATION_MESSAGE); 
                          }
                          else
                              JOptionPane.showMessageDialog(null,"This image is not supported!","Warning!",JOptionPane.WARNING_MESSAGE);
                              
                  	  }		
                  	  else{	//input is a directory!
                  		  File[] content = input.listFiles(new JpegExtensionFilter());
                  		  boolean flagWarning = false; //If true one or more images are not supported!
                          boolean flagPullDown = false; //There is at least a valid image
                          for(int i=0;i<content.length;i++){
                  			  String exif = readMetaData(content[i]);
                  			  psnr = extractThumbnail(content[i]);
                  			  if(psnr != -2){
                                  comboBoxFileList.addItem(content[i]);
                                  writeReport(content[i], exif, psnr);
                                  flagPullDown = true;
                              }
                              else
                                  flagWarning = true;
                  		  }
                          if (flagWarning)
                              JOptionPane.showMessageDialog(null,"One or more images are not supported!","Warning!",JOptionPane.WARNING_MESSAGE);
                          if (flagPullDown)
                              JOptionPane.showMessageDialog(null,"Select an image from the pull-down menu","Done!",JOptionPane.INFORMATION_MESSAGE); 
                        }
                    }
                    catch(Exception e)
                    {
                    	JOptionPane.showMessageDialog(null,"You should select something first...","Warning!",JOptionPane.WARNING_MESSAGE);
                    }
               }
			}
		});
		buttonBrowse.setBounds(104, 43, 256, 29);
		frmExifViewer.getContentPane().add(buttonBrowse);
		
		//Separator 1
		JSeparator separator1 = new JSeparator();
		separator1.setBounds(10, 73, 444, 14);
		frmExifViewer.getContentPane().add(separator1);
		
		//Separator 2
		JSeparator separator2 = new JSeparator();
		separator2.setBounds(10, 355, 444, 14);
		frmExifViewer.getContentPane().add(separator2);
				
		//Separator 3
		JSeparator separator3 = new JSeparator();
		separator3.setBounds(10, 141, 444, 14);
		frmExifViewer.getContentPane().add(separator3);
		
		//Separator 4
		JSeparator separator4 = new JSeparator();
		separator4.setBounds(10, 410, 444, 14);
		frmExifViewer.getContentPane().add(separator4);
		
		//Label Browse
		JLabel labelBrowse = new JLabel("Select an Image/Directory:");
		labelBrowse.setBounds(10, 11, 280, 21);
		frmExifViewer.getContentPane().add(labelBrowse);
        
		//Label Thumbnail
		JLabel textThumbnail = new JLabel("Image Thumbnail:");
		textThumbnail.setBounds(10, 152, 280, 21);
		frmExifViewer.getContentPane().add(textThumbnail);
		
		//Label Exif
		JLabel labelExif = new JLabel("Exif Data:");
		labelExif.setBounds(10, 421, 280, 21);
		frmExifViewer.getContentPane().add(labelExif);
		
		//Label FileList
		JLabel labelFileList = new JLabel("File List:");
		labelFileList.setBounds(10, 83, 280, 21);
		frmExifViewer.getContentPane().add(labelFileList);
		
		//Label PSNR
		JLabel labelPSNR = new JLabel("PSNR:");
		labelPSNR.setBounds(10, 378, 44, 21);
		frmExifViewer.getContentPane().add(labelPSNR);
		
		

		imp = WindowManager.getCurrentImage();
		if(imp != null){
			String filePath = imp.getOriginalFileInfo().directory+"/"+imp.getTitle();
  		  	File file = new File(filePath);
			comboBoxFileList.setEnabled(true);
         	comboBoxFileList.addItem(file);
         	String exif = readMetaData(file);
         	psnr = extractThumbnail(file);
         	writeReport(file, exif, psnr);
         	comboBoxFileList.setSelectedIndex(1);

		}
		
	}
	
	
	/** Extract exif data from report file and fill the exif table **/
	private void fillExifTable(File file){
		
		int count = myModel.getRowCount();
		if(count!=0)
			for(int i=0;i<count;i++)
				myModel.removeRow(0);
		
		BufferedReader in = null;
		
		try{	
			in = new BufferedReader(new FileReader(file.getParent()+"/Reports/"+file.getName()+".txt"));
			
			//Extract Exif Data
			String s = null;
			while(!(s=in.readLine()).equals("")){
				
				int i=0;
				String name="";
				while(s.charAt(i)!=':')	
					name+=(""+s.charAt(i++));
				
				String value="";
				while(++i<s.length())
					value+=(""+s.charAt(i));
				
				String[] row = {name,value};
				myModel.addRow(row);
			}
			
			//Extract PSNR
			in.readLine();
			s = in.readLine();
			int i=0;
			while(s.charAt(i)!=':')
				i++;
			psnr = Double.parseDouble(s.substring(i+2,s.length()));
			
		}catch(Exception e){
			e.printStackTrace();
		}
	}
	
	
	/** Extract thumbnail from image exif data and calculate PSNR between the thumbnail and a resized copy of the image **/
	private static double extractThumbnail(File file) {	

        IImageMetadata metadata = null;
        byte[] thumb = null;				
        BufferedImage thumbImage = null;
        BufferedImage originalImage = null;
        BufferedImage resizedImage = null;
        
		try {
            metadata = Sanselan.getMetadata(file);
        } catch (ImageReadException e) {
            e.printStackTrace();
        } catch (IOException e) {
            e.printStackTrace();
        }

		//If file is a jpeg image...
        if (metadata instanceof JpegImageMetadata) {
            JpegImageMetadata jpegMetadata = (JpegImageMetadata) metadata;
            ArrayList dirs = jpegMetadata.getExif().getDirectories();
            boolean thumbFound = false;

            for (int i=0; i<dirs.size(); i++) {
            	TiffImageMetadata.Directory tiffdir = (TiffImageMetadata.Directory) dirs.get(i);   	
               	if( tiffdir.getJpegImageData() != null ){
            		thumb = tiffdir.getJpegImageData().data;   		
            		if( thumb != null ){
            			try {
            				InputStream in = new ByteArrayInputStream(thumb);
							thumbImage = ImageIO.read(in);
							thumbFound = true;					
            			} catch (IOException e) {
							e.printStackTrace();
						}	
            		}	 
            	}
            }
			if (!thumbFound)
           		return -1;
            
            try{
            	originalImage = ImageIO.read(file);
            }
            catch(IOException e){
            	e.printStackTrace();
            }
            
			//Rescale the original image 
			resizedImage = new BufferedImage(thumbImage.getWidth(),thumbImage.getHeight(),BufferedImage.TYPE_INT_RGB);
			Graphics2D graphics2D = resizedImage.createGraphics();
			graphics2D.setRenderingHint(RenderingHints.KEY_INTERPOLATION,RenderingHints.VALUE_INTERPOLATION_BILINEAR);
			graphics2D.drawImage(originalImage, 0, 0, thumbImage.getWidth(), thumbImage.getHeight(), null);
	
			//Save the thumbanail
			try {
				new File(file.getParent()+"/Thumbnails").mkdir();
				ImageIO.write(thumbImage, "jpg",new File(file.getParent()+"/Thumbnails/"+file.getName().substring(0, file.getName().length()-4)+"Thumb.jpg"));
			} catch (IOException e) {
				e.printStackTrace();
			}
		}
        else{
                //JOptionPane.showMessageDialog(null,"This image is not supported!","Warning!",JOptionPane.WARNING_MESSAGE);
            return -2;
        }
        return calculatePSNR(resizedImage,thumbImage); 
	}
	
	
	/** Calculate PSNR between the thumbnail and a resized copy of the image **/
    private static double calculatePSNR(BufferedImage resizedImage, BufferedImage thumbImage) {
	
    	int width = resizedImage.getWidth();
    	int height = resizedImage.getHeight();
    	int[][][] resized;
    	int[][][] thumb;
    	double mseR=0;
    	double mseG=0;
    	double mseB=0;
    	double psnrR=0;
    	double psnrG=0;
    	double psnrB=0;
    	
    	resized = imageToRGB(resizedImage);
    	thumb = imageToRGB(thumbImage);
    	
    	//MSEs
		for(int i=0;i<height;i++)
			for(int j=0;j<width;j++){
				mseR += Math.pow(resized[i][j][0] - thumb[i][j][0], 2);
				mseG += Math.pow(resized[i][j][1] - thumb[i][j][1], 2);
				mseB += Math.pow(resized[i][j][2] - thumb[i][j][2], 2);
			}
		
		mseR /=(width*height);
		mseG /=(width*height);
		mseB /=(width*height);
		
		psnrR = -10*Math.log10(mseR/(255*255));
		psnrG = -10*Math.log10(mseG/(255*255));
		psnrB = -10*Math.log10(mseB/(255*255));
		
		return (psnrR + psnrG + psnrB)/3;
	}

    
   
    /** Get the 3 color planes from the BufferedImage  **/
    private static int[][][] imageToRGB(BufferedImage image){

    	int width = image.getWidth();
    	int height = image.getHeight();
    	int[][][] rgb = new int[height][width][3];
    			
    	for(int i=0;i<height;i++)
    		for(int j=0;j<width;j++){
    			int pixel = image.getRGB(j, i);
    			rgb[i][j][0] = (pixel & 0xff0000) >> 16;	//RED;
    			rgb[i][j][1] = (pixel & 0x00ff00) >> 8;		//GREEN;
            	rgb[i][j][2] = (pixel & 0x0000ff);			//BLUE;
    		} 			
    	return rgb;
	}
    

    /** Write a report file with exif data and psnr **/
	private void writeReport(File input, String exif,double psnr){
		try {
			new File(input.getParent()+"/Reports").mkdir();
			FileOutputStream file = new FileOutputStream(input.getParent()+"/Reports/"+input.getName()+".txt");
			PrintStream out = new PrintStream(file);
			
            if(psnr == -1)
                out.print(exif+"\n\nImage doesn't contain any thumbnail!");
			else
                out.print(exif+"\n\nPSNR between rescaled image and thumbnail is: "+psnr);
			out.close();
			
		} catch (FileNotFoundException e) {
			e.printStackTrace();
		}
		
	}
    
 
	/** Read metadata from image file and return them in a string **/
    private static String readMetaData(File file) {
    	
        IImageMetadata metadata = null;
        StringBuffer s = new StringBuffer();	//Output
        
        try {
            metadata = Sanselan.getMetadata(file);
        } catch (ImageReadException e) {
            e.printStackTrace();
        } catch (IOException e) {
            e.printStackTrace();
        }

        //If file is a jpeg image...
        if (metadata instanceof JpegImageMetadata) {
        	
            JpegImageMetadata jpegMetadata = (JpegImageMetadata) metadata;
                
            //GPS data
            exifMetadata = jpegMetadata.getExif();
            if (exifMetadata != null) {
                try {
                    TiffImageMetadata.GPSInfo gpsInfo = exifMetadata.getGPS();
                    if (gpsInfo != null) {
                        double longitude = gpsInfo.getLongitudeAsDegreesEast();
                        double latitude = gpsInfo.getLatitudeAsDegreesNorth();
                        
                        s.append("GPS Description: " + gpsInfo+"\n");
                        s.append("GPS Longitude (Degrees East): "+longitude+"\n");
                        s.append("GPS Latitude (Degrees North): "+latitude+"\n");    
                    }
                } 
                catch (ImageReadException e) {
                    e.printStackTrace();
                }
            }
            
            //Exif data
            ArrayList items = exifMetadata.getItems(); 
            for (int i=0; i<items.size();i++) 
            	s.append(items.get(i)+"\n");
            
        }
        
        return s.toString();
    }
}



/** This class implements a filter for jpeg files **/
class JpegExtensionFilter implements FilenameFilter{
	
	private String ext1 =".jpeg";
	private String ext2 =".JPEG";
	private String ext3 =".jpg";
	private String ext4 =".JPG";

	public boolean accept(File dir, String name){
		if (name.endsWith(ext1) || name.endsWith(ext2) || name.endsWith(ext3) || name.endsWith(ext4))
			return true;
		return false;
	}
}

/** This class define the table model with not-editable cells **/
class myTableModel extends DefaultTableModel
{

	static String[] labels = {"Name","Value"};
	static String[][] content = {};
	
	public myTableModel( ){
		super(content,labels);
	 }
	
	public boolean isCellEditable(int row,int cols){
			return false;                                                                                   
    }
}         